///////////////////////////////////////////////////////////////
// The PC Benchmark (pcb)
// (c) 2015, Marko Oette (www.oette.info)
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "plattform.h"
#include "pcBench.h"
#include "tests.h"

#include <time.h>

long testCycles = TEST_CYCLES;			// number of tests to perform
char *destPath = NULL;					// drive test destination
ofstream logStream;							// logfile output stream

// Implementation
int main(int argc, char* argv[])
{
	// Init random number generator
	time_t t;
    time(&t);
    srand((unsigned int)t);

	// Get the number of test cycles from the cmd line
	char *dummy;
	bool goInteractive = true;
	unsigned int threads = 1;

	if( argc >= 2 )	// Second argument is expected to be a cry for help or the number of test cycles.
	{
		if( strcmp(argv[1], "-h" ) == 0 || strcmp(argv[1], "/h" ) == 0 )
		{
			printf("Usage: %s [cycles] [threads] [logfile] [destpath]\nNo commands start pcBench in interactive single core mode.", argv[0] );
			return 0;
		}
		testCycles = strtol(argv[1], &dummy, 10);
		// goInteractive = false;
	}

	if (argc >= 3)	// Second argument is expected to be the number of threads to use
	{
		threads = strtol(argv[2], &dummy, 10);
		// goInteractive = false;
	}
	
	if( argc >= 4 )	// Third argument is the name of the logfile
	{
		logStream.open(argv[3], ios::out | ios::trunc);
		if( logStream.bad() || !logStream.is_open() )
		{
			cout << "Can not open log file: " << argv[3] << "!\n";;
			return -5;
		}
		else
			cout << "Logging to file: " << argv[3] << ".\n\n";

		goInteractive = false;
	}
	
	if( argc >= 5 )	// Third argument is the name of the logfile
	{
		destPath = argv[4];
		goInteractive = false;
	}

	cout << "*********************************************************************\n";    
	cout << "*                      PC Benchmark v. 1.0.1                        *\n";
	cout << "*             (c) 2015 Marko Oette - www.oette.info                 *\n";
	cout << "*********************************************************************\n\n";
	cout << "Compiled in " << TOSTRING(BITNESS) << " bits by " << TOSTRING(__VERSION__) << " flavor: " << TOSTRING(PCBVER) << "\n";	
	//cout << "Compiled on: " <<  TOSTRING(__TIMESTAMP__) << "\n";
	
	if( logStream.good() )
	{
		logStream << "PC Benchmark v. 1.0.1 (c) 2015 Marko Oette, www.oette.info\n";
		logStream << "Compiled in " << TOSTRING(BITNESS) << " bits by " << TOSTRING(__VERSION__) << " flavor: " << TOSTRING(PCBVER) << "\n";	
   		//logStream << "Compiled on: " <<  TOSTRING(__TIMESTAMP__) << "\n";
		logStream << "Using " << dec << threads << " thread(s) and " << testCycles << " cycles.\n";
		logStream << "##SYS;" << TOSTRING(BITNESS) << ";" << TOSTRING(PCBVER) << "\n";
		// printSysInfo(logStream);
	}

	int bits = VerifySizeType();
	cout << "SYS-CHECK: size_t is " << dec << bits << " bits, should be " << dec << BITNESS << " - ";
	if( bits != BITNESS )
	{
		cout << "Aborting.";
		return -1;
	}
	else
		cout << "OK.";
	cout << "\n";

	cout << "Using " << dec << threads << " thread(s) and " << testCycles << " cycles.\n";

	if( !goInteractive )	// command line sets goInteractive to false
	{
		// Perform all testing...
		RunCpuTests(threads);
		RunMemTests(threads);

		if( destPath != NULL )
			RunDiskTests();

		return 0;
	}

	// Enter menu
	while( goInteractive )
	{
		cout << "\nMENU - Available options\n";
		cout << "  c - Test CPU speed\n";
		cout << "  m - Measure memory speed (read+write, memcpy)\n";
		cout << "  b - Measure memory bandwidth (write only, memset)\n";
		cout << "  d - Test disk speed\n";
		cout << "  a - Perform all tests\n";
		cout << "\n";
		cout << "  Anything else - Exit\n\n";
		cout << "Your choice: ";
		cin.sync();
	
		char key = 0;
		cin.get( key );
		while(cin.peek() != '\n' )
			cin.ignore();	// Ignore the next character. (Until Return)
		cin.ignore();	// Ignore the return as well.
		cout << "\n";

		switch( key )
		{
			case 'c':
				RunCpuTests(threads);
				break;
			case 'm':
				RunMemTests(threads);
				break;
			case 'b':
				MeasureMemoryBandwith(threads);
				break;
			case 'd':
				RunDiskTests();
				break;
			case 'a':
				RunCpuTests(threads);
				MeasureMemoryBandwith(threads);
				RunMemTests(threads);
				RunDiskTests();
				break;
			default:
				goInteractive = false;
		}
	}
	
	cout << "\n\nBye.\n";
	
	return 0;
}

/**
	Returns the SI prefix and responding SI value for the given integer value.
*/
/* void SIPrefix(INT_TYPE val, char &prefix, INT_TYPE &siVal)
{	
	int id = 1;	// The values we deal with are already KILO
	
	while(val > 1000)
	{
		val /= 1000;
		id++;
	}                
	
	switch(id)
	{
	case 1:
		prefix = 'k';
		break;
	case 2:
		prefix = 'M';
		break;
	case 3:
		prefix = 'G';
		break;
	case 4:
		prefix = 'T';
		break;
	case 5:
		prefix = 'P';
		break;
	case 6:
		prefix = 'E';
		break;
	default:
		// Do nothing;
		break;
	}
	
	siVal = val;
}*/

/**
	Test how large size_t is.
*/
unsigned int VerifySizeType()
{
	size_t max  = 0;
	size_t cur = 1;
	unsigned int bits = 0;

	while(true)
	{
		if( max < cur )
		{
			max = cur;
			bits++;
			cur = cur << 1;
		}
		else
		{
			// overflow
			break;
		}
	}
	
	return bits;
}

/**
	Performs the actual testing.
*/	
void RunCpuTests(unsigned int threads)
{
	// Go Realtime
	if(!setProcessPriority(PROC_PRIO_REAL))
	{
		cout<< "WARNING: Can not go realtime! Trying high priority.\n";
		if (!setProcessPriority(PROC_PRIO_HIGH))
		{
			cout << "WARNING: Can not set process to high priority! Results will be inaccurate.\n";
			logStream << "WARNING: Can not set process to high priority! Results will be inaccurate.\n";
		}
	}
		

	// Calibrate speed
	LNG_TYPE c = 0;
	LNG_TYPE ops = 1;
	LNG_TYPE speed = 0;

	cout << "\nCalibrating CPU Test...\n";
	while( c < 1000 )
	{
		ops *= 10;
		c = timedTest( Calibrate, ops, 0, NULL );
		
		cout << "#";
		clog << "";
	}
	speed = (ops / ((double)c));	
	speed *= 1000; // ms->Seconds

	cout << "\r  Done. (Result: CPU does " << speed << " OPS per second on one core.)\n";

	
	// ***************


	cout << "\nTesting CPU...\n";
	logStream  << "\nTesting CPU...\n";
                                    	
	int i = 0;
	FP_TYPE time = 0;
	for( i = 0; i < testCycles; i++)
	{
		cout << "\r  Measuring CPU integer operation speed (" << dec << testCycles << " cycles) " << i + 1 << "/" << testCycles;
		clog << "";

		FP_TYPE tt = 0;
		if (threads>1)
			tt = timedMTTest(threads, MeasureIOPS, speed, 0, NULL);
		else
			tt = timedTest(MeasureIOPS, speed, 0, NULL);

		assert( tt > 0 );
		time += tt;

		if (i > 0)
			time /= 2;
	}

	if( time <= 0 )
	{
		cout << "\nSeems your machine is too fast to measure using this tool.\n";
		setProcessPriority(PROC_PRIO_NORMAL);
		return;
	}
	
	LNG_TYPE iops = (speed / time)*1000;
	cout << "\r  CPU integer speed is:\t\t" << threads << "x" << dec << iops << " IOPS.                \n";
	logStream  << "  CPU integer speed is:\t\t" << threads << "x" << dec << iops << " IOPS.\n";
	
	
	/// *********
	 
	 
    time = 0;
	for( i = 0; i < testCycles; i++)
	{
		cout << "\r  Measuring CPU floating point speed (" << dec << testCycles << " cycles) " << i + 1 << "/" << testCycles;
		clog << "";

		FP_TYPE tt = 0;
		if (threads>1)
			tt = timedMTTest(threads, MeasureFLOPS, speed, 0, NULL);
		else
			tt = timedTest(MeasureFLOPS, speed, 0, NULL);

		assert (tt > 0 );
		time += tt;

		if (i > 0)
			time /= 2;
	}

	if( time == 0 )
	{
		cout << "\nSeems your machine is too fast to measure using this tool.\n";
		setProcessPriority(PROC_PRIO_NORMAL);
		return;
	}

	LNG_TYPE flops = (speed / time) * 1000;
	cout << "\r  CPU floating point speed is:\t" << threads << "x" << dec << flops  << " FLOPS.                \n";
	logStream  << "  CPU floating point speed is:\t" << threads << "x" << dec << flops  << " FLOPS.\n";
	
	
	/// *********
	 	 
    time = 0;
	for( i = 0; i < testCycles; i++)
	{
		cout << "\r  Measuring CPU bit shifting speed (" << dec << testCycles << " cycles) " << i + 1 << "/" << testCycles;
		clog << "";

		FP_TYPE tt = 0;
		if (threads>1)
			tt = timedMTTest(threads, MeasureBitShifting, speed, 0, NULL);
		else
			tt = timedTest(MeasureBitShifting, speed, 0, NULL);

		assert (tt > 0 );
		time += tt;

		if (i > 0)
			time /= 2;
	}

	if( time <= 0 )
	{
		cout << "\nSeems your machine is too fast to measure using this tool.\n";
		setProcessPriority(PROC_PRIO_NORMAL);
		return;
	}

	LNG_TYPE bsops = (speed / time)*1000;
	cout << "\r  CPU bit shifting speed is:\t" << threads << "x" << dec << bsops << " BSOPS.               \n";
	logStream  << "  CPU bit shifting speed is:\t" << threads << "x" << dec << bsops << " BSOPS.\n";


	/// *********
	 	 
	time = 0;
	for( i = 0; i < testCycles; i++)
	{
		cout << "\r  Measuring CPU math speed (" << dec << testCycles << " cycles) " << i + 1 << "/" << testCycles;
		clog << "";

		// Doing speed / 10 operations as maths are slower...
		FP_TYPE tt = 0;
		if (threads>1)
			tt = timedMTTest(threads, MeasureFloatingMath, speed / 10, 0, NULL);
		else
			tt = timedTest(MeasureFloatingMath, speed / 10, 0, NULL);

		assert (tt > 0 );
		time += tt;

		if (i > 0)
			time /= 2;
	}

	if( time <= 0 )
	{
		cout << "\nSeems your machine is too fast to measure using this tool.\n";
		setProcessPriority(PROC_PRIO_NORMAL);
		return;
	}

	LNG_TYPE mtops = ((speed/10) / time)*1000;
	cout << "\r  CPU math speed is:\t\t" << threads << "x" << dec << mtops << " Math-OPS.               \n";
	logStream  << "  CPU math speed is:\t\t" << threads << "x" << dec << mtops << " Math-OPS.\n";

	if (logStream.good())
	{
		strstream retCodes;
		logStream << "\nMachine readable results: \"CPU\";THREADS;IOPS;FLOPS;BSOPS;MATH-OPS;CHECKSUM\n";
		retCodes << "##CPU;" << dec << threads << ";" << dec << iops << ";" << dec << flops << ";" << dec << bsops << ";" << dec << mtops << ";" << ends;
		char* szRetCode = retCodes.str();
		unsigned long crc = getStringHash(szRetCode);
		logStream << szRetCode << hex << crc << "\n";
	}

	setProcessPriority(PROC_PRIO_NORMAL);
}

void RunMemTests(unsigned int threads)
{
	// Go Realtime
	if(!setProcessPriority(PROC_PRIO_REAL))
	{
		cout<< "WARNING: Can not go realtime! Trying high priority.\n";
		if (!setProcessPriority(PROC_PRIO_HIGH))
		{
			cout << "WARNING: Can not set process to high priority! Results will be inaccurate.\n";
			logStream << "WARNING: Can not set process to high priority! Results will be inaccurate.\n";
		}
	}

	LNG_TYPE ops = 1;
	cout << "\nTesting memory speed (read+write, memcpy)...\n";
	logStream  << "\nTesting memory speed (read+write, memcpy)...\n";

	INT_TYPE b = 0;
	INT_TYPE c = 0;
	INT_TYPE maxBlock = 0;
	INT_TYPE maxSpeed = 0;
	LNG_TYPE maxOps = 0;
	strstream memcpySpdDetails;
	memcpySpdDetails << "##SPD;" << dec << threads << ";";

	for( b=1; b<=1024; b*=2 )
	{
		cout << "  " << b << " KiBytes: ";
		clog << "";             // Forces flushing cout
		c = 0;
		ops = 1;
		FP_TYPE mcspeed = 0;
		while( c < 1000 )
		{
			ops *= 10;	// Increase the operations counter by factor 10

			if (threads > 1)
				c = timedMTTest(threads, ApproximateMemCpy, ops, b, NULL);
			else
				c = timedTest(ApproximateMemCpy, ops, b, NULL);

			cout << "#";
			clog << "";
		}
		mcspeed = (ops / ((double)c));	

		size_t bufSize = b * 1024;

		// Remember the fastest block size
		if (b*mcspeed > maxSpeed)
		{
			maxBlock = b;
			maxSpeed = b*mcspeed;
			maxOps = ops;
		}

		cout << "\r  Speed @ "<< b << " KiB: " << dec << threads << "x" << /* dec << mcspeed << " K-OPS / " <<*/ b*mcspeed << " MiB/s.\n";
		clog << "";             // Forces flushing cout

		logStream  << "  Speed @ "<< b << " KiB: " << dec << threads << "x" << /* dec << mcspeed << " K-OPS / " <<*/ b*mcspeed << " MiB/s.\n";
		
		// Preserve machine readable
		memcpySpdDetails << dec << b*mcspeed << ";";
	}


	/// **************

	cout << "\n  Best block / speed was (" << dec << maxSpeed << " MiB/s @ " << maxBlock << " KiB blocks)\n";
	
	FP_TYPE time = 0;
	for (int i = 0; i < testCycles; i++)
	{
		cout << "\r  Stress-Test: Repeating in " << dec << testCycles << " cycles with " << maxBlock << " KiB blocks. " << i+1 << "/" << testCycles;
		clog << "";

		FP_TYPE tt = 0;
		if (threads>1)
			tt = timedMTTest(threads, ApproximateMemCpy, maxOps, maxBlock, NULL);
		else
			tt = timedTest(ApproximateMemCpy, maxOps, maxBlock, NULL);

		assert(tt > 0);
		time += tt;
		if (i > 0)
			time /= 2;
	}

	if (time <= 0)
	{
		cout << "\nSeems your machine is too fast to measure using this tool.\n";
		setProcessPriority(PROC_PRIO_NORMAL);
		return;
	}

	LNG_TYPE mcpyops = ((maxOps) / time) * maxBlock;
	cout << "\r  The max. memory speed (memcpy) is: " << threads << "x" << dec << (mcpyops) << " MiB/s @ " << maxBlock << " KiB blocks.           \n";
	logStream << "  The max. memory speed (memcpy) is: " << threads << "x" << dec << (mcpyops) << " MiB/s @ " << maxBlock << " KiB blocks.\n";

	if (logStream.good())
	{
		logStream << "\nMachine readable results: \"SPD\";THREADS;11x{MiB/s;1-1024K};CHECKSUM\n";

		memcpySpdDetails << ends;
		char* szSpdDetail = memcpySpdDetails.str();
		unsigned long crc = getStringHash(szSpdDetail);
		logStream << szSpdDetail << hex << crc << "\n";
		memcpySpdDetails.rdbuf()->freeze(false);		// Clean up

		logStream << "\nMachine readable results: \"MEM\";THREADS;MCPYBLCK;MCPYSPD;CHECKSUM\n";
		strstream retCodes;
		retCodes << "##MEM;" << dec << threads << ";" << dec << maxBlock << ";" << dec << mcpyops << ";" << ends;
		char* szRetCode = retCodes.str();
		crc = getStringHash(szRetCode);
		logStream << szRetCode << hex << crc << "\n";
		retCodes.rdbuf()->freeze(false);				// Clean up
	}

	setProcessPriority(PROC_PRIO_NORMAL);
	return;
}

void MeasureMemoryBandwith(unsigned int threads)
{
	// Go Realtime
	if (!setProcessPriority(PROC_PRIO_REAL))
	{
		cout << "WARNING: Can not go realtime! Trying high priority.\n";
		if (!setProcessPriority(PROC_PRIO_HIGH))
		{
			cout << "WARNING: Can not set process to high priority! Results will be inaccurate.\n";
			logStream << "WARNING: Can not set process to high priority! Results will be inaccurate.\n";
		}
	}
	
	unsigned int b = 1;		// Start block size in bytes

	cout << "\nTesting memory bandwith (writing, memset)...\n";
	logStream << "\nTesting memory bandwith (writing, memset)...\n";

	strstream memcpySpdDetails;
	memcpySpdDetails << "##MST;" << dec << threads << ";";

	LNG_TYPE max = 0;
	LNG_TYPE maxBlock = 0;

	while( b <= (1024 * 2048) )		// Maximum block size in bytes
	{
		LNG_TYPE tt = 0;
		LNG_TYPE ops = 1;
		LNG_TYPE speed = 0;
		
		cout << "  " << b << " bytes: ";
		clog << "";             // Forces flushing cout
		while (tt < 1000)
		{
			ops *= 10;
			if (threads>1)
				tt = timedMTTest(threads, ApproximateMemSet, ops, b, NULL);
			else
				tt = timedTest(ApproximateMemSet, ops, b, NULL);
			
			cout << "#";
			clog << "";
		}

		speed = (ops * b / tt);
		speed *= 1000;	// ms->Seconds
		INT_TYPE val = (speed/1024)/1024;

		if( val > max )
		{
			max = val;
			maxBlock = b;
		}

		cout << "\r  Bandwidth @ " << b << " bytes: " << val << " MiB/s.         \n";
		logStream << "  Bandwidth @ " << b << " bytes: " << val << " MiB/s.\n";
		
		// Preserve machine readable
		memcpySpdDetails << dec << val << ";";
		
		b*=2;
	}

	setProcessPriority(PROC_PRIO_NORMAL);

	char* szSpdDetail = memcpySpdDetails.str();
	unsigned long crc = getStringHash(szSpdDetail);
	logStream << szSpdDetail << hex << crc << "\n";
	memcpySpdDetails.rdbuf()->freeze(false);		// Clean up

	cout << "\nThe max. memory bandwidth (memset) is: " << threads << "x" << max << " MiB/s @ " << maxBlock << " byte blocks.\n";
	logStream << "\nThe max. memory bandwidth (memset) is: " << threads << "x" << max << " MiB/s @ " << maxBlock << " byte blocks.\n";
	

	return;
}

void RunDiskTests()
{
	// Go High Priority (Realtime is no good idea as disk access will be reduced.)
	if(!setProcessPriority(PROC_PRIO_HIGH))
	{
		cout<< "WARNING: Can not set process to high priority! Results will be inaccurate.\n";
		logStream << "WARNING: Can not set process to high priority! Results will be inaccurate.\n";
	}

	INT_TYPE ops = 1;
	INT_TYPE c = 0;
	INT_TYPE b = 0;

	while( destPath == NULL )
	{
		cout << "\nDisk-Spped-Test \n";
		cout << "Please enter a valid path for testing: (eg: C:, B: - empty aborts)\n";
		cin.sync();		// Flush anything the user wrote 

		char *buf = new char[65];
		cin.getline( buf, 64 );
		
		strstream sTest;
		sTest << buf;
		delete buf; buf=0;

		if( sTest.pcount() == 0 )
			return;

		sTest << "\\test.io" << ends;

		char *szTest = sTest.str();
		FILE* f = fopen(szTest, "wb"); 
		if( f == NULL )
			cout << "\nThis path is invalid. (Or we do not have write permissions!)\n";
		else
		{
			fclose(f);
			destPath = szTest;	// Freeing is to be done otherwhere!
			break;
		}

		//Free it as it was useless
		free(szTest);
	}

	cout << "\nTesting disk speed (1 - 1024 KiB blocks)...\n";
	logStream  << "\nTesting disk speed (1 - 1024 KiB blocks) on path: " << destPath << "\n";

	strstream sTest;
	sTest << destPath;
	sTest << "\\test.io" << ends;
	char *szTest = sTest.str();

	for( b=1; b<=1024; b*=2 )
	{
		cout << "  Testing with block size = " << b << " KiBytes... ";
		clog << "";             // Forces flushing cout
		c = 0;
		ops = 1;
		FP_TYPE dspeed = 0;
		while( c < 1000 )
		{
			ops *= 10;	// Increase the operations counter by factor 10
			c = timedTest( ApproximateDiskWriteSpeed, ops, b, szTest );
			cout << "#";
			clog << "";
		}
		dspeed = (ops / ((double)c));	

		size_t bufSize = b * 1024;

		cout << "\r  Speed @ "<< b << " KiB: " << dspeed << " K-OPS / " << (bufSize/1024)*dspeed << " MiB per second.             \n";
		logStream  << "  Speed @ "<< b << " KiB: " << dspeed << " K-OPS / " << (bufSize/1024)*dspeed << " MiB per second.             \n";
		clog << "";             // Forces flushing cout
	}

	free( szTest );
}

unsigned long getStringHash(char* str)
{
	unsigned int h = 0;
	if (h == 0 && strlen(str) > 0)
	{
		int len = strlen(str);
		for (int i = 0; i < len; i++)
		{
			h = 31 * h + str[i];
		}
	}
	return h;
}