///////////////////////////////////////////////////////////////
// The PC Benchmark (pcb)
// (c) 2015, Marko Oette (www.oette.info)
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "plattform.h"
#include <time.h>


/**
	Thread Data structure
*/
typedef struct ThreadDataStruct 
{
	int(*pfn)(LNG_TYPE, LNG_TYPE, const void*);	// Function pointer to be called
	LNG_TYPE p1;		// Parameter 1 (optional)
	LNG_TYPE p2;		// Parameter 2 (optional)
	const void* buf;	// Parameter 3 (void pointer) (optional)

	LNG_TYPE ret;		// Return value (optional)
} ThreadData;

/**
	Thread Function for multithreaded testing.
*/
DWORD WINAPI multiThreadRunner(LPVOID lpParam)
{
	ThreadData *pData = (ThreadData*)lpParam;
	
	pData->ret = timedTest(pData->pfn, pData->p1, pData->p2, pData->buf);
	return 0;
}

/**
	Measures the time in milli seconds required to perform the given function.
	Accuracy depends on the target plattform.
*/
LNG_TYPE timedMTTest(unsigned int threads, int(*pfn)(LNG_TYPE, LNG_TYPE, const void*), LNG_TYPE p1, LNG_TYPE p2, const void* buf)
{
	LNG_TYPE t = 0;
	unsigned int i = 0;
	 
	// Create threads
	HANDLE  *hThreadArray = new HANDLE[threads]; // (HANDLE*)malloc(threads * sizeof(HANDLE));
	DWORD   *dwThreadIdArray = new DWORD[threads]; // (DWORD*)malloc(threads * sizeof(DWORD));
	ThreadData *tData = new ThreadData[threads]; // (ThreadData*)malloc(threads * sizeof(ThreadData));

	for (i = 0; i < threads; i++)
	{

		tData[i].pfn = pfn;
		tData[i].p1 = p1;
		tData[i].p2 = p2;
		tData[i].buf = buf;

		hThreadArray[i] = CreateThread(
			NULL,                   // default security attributes
			0,                      // use default stack size  
			multiThreadRunner,       // thread function name
			&tData[i],				// argument to thread function 
			0,                      // use default creation flags 
			&dwThreadIdArray[i]);   // returns the thread identifier 
	}

	// Wait for threads
	WaitForMultipleObjects(threads, hThreadArray, TRUE, INFINITE);

	// Clean up
	for (i = 0; i < threads; i++)
	{
		CloseHandle(hThreadArray[i]);
		t += tData[i].ret;

		if( i > 0 )	// Average the time over the threads
			t /= 2;
	}
	
	delete hThreadArray;
	delete dwThreadIdArray;
	delete tData;

	return t;
}

/**
	Measures the time in milli seconds required to perform the given function.
	Accuracy depends on the target plattform.
*/
LNG_TYPE timedTest( int(*pfn)( LNG_TYPE, LNG_TYPE, const void*), LNG_TYPE p1, LNG_TYPE p2, const void* buf )
{
	// Get start time
#ifdef _WIN32
	#ifdef _USE_W32PERFCOUNTER
		#pragma message ("_USE_W32PERFCOUNTER is set.")
		LARGE_INTEGER StartingTime, EndingTime, ElapsedMicroseconds;
		LARGE_INTEGER Frequency;

		QueryPerformanceFrequency(&Frequency); 
		QueryPerformanceCounter(&StartingTime);
	#else
		DWORD dwStart = ::GetTickCount();
	#endif
#endif

	// Perform test
	int ret = pfn( p1, p2, buf );

	// Get duration
#ifdef _WIN32
	#ifdef _USE_W32PERFCOUNTER
		QueryPerformanceCounter(&EndingTime);
		ElapsedMicroseconds.QuadPart = EndingTime.QuadPart - StartingTime.QuadPart;

		//
		// We now have the elapsed number of ticks, along with the
		// number of ticks-per-second. We use these values
		// to convert to the number of elapsed microseconds.
		// To guard against loss-of-precision, we convert
		// to milliseconds *before* dividing by ticks-per-second.
		//
		ElapsedMicroseconds.QuadPart *= 1000000;
		ElapsedMicroseconds.QuadPart /= Frequency.QuadPart;
		
		if( ret == -1 )
			return 0;

		return (LNG_TYPE)(ElapsedMicroseconds.QuadPart / 1000);
	#else
		DWORD dwDuration = ::GetTickCount() - dwStart;

		if( ret == -1 )
			return 0;

		return dwDuration;
	#endif
#endif
}

/**
	This function ensures the benchmark is running in highest possible priority...
*/
bool setProcessPriority( int prio )
{
	bool ret = false;
#ifdef _WIN32
	switch( prio )
	{
		case PROC_PRIO_REAL:
			if(SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS))
				ret = true;	
			break;
		case PROC_PRIO_HIGH:
			if(SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS))
				ret = true;	
			break;
		default:
			if(SetPriorityClass(GetCurrentProcess(), NORMAL_PRIORITY_CLASS))
				ret = true;	
			break;
	}

#endif
	return ret;
}

void printSysInfo(ofstream &log)
{
	SYSTEM_INFO info; 
	GetSystemInfo(&info);
}
