///////////////////////////////////////////////////////////////
// The PC Benchmark (pcb)
// (c) 2015, Marko Oette (www.oette.info)
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "plattform.h"
#include "tests.h"

#include <math.h>

/**
	Estimates the raw speed of the CPU
	ops is the number of operations to perform
*/
int Calibrate(LNG_TYPE ops, LNG_TYPE dummy, const void* buf)
{
	LNG_TYPE c = ops;
	
	while(c --> 0)
	{		
	}		
	
	return 0;
}

/**
	Halts the current thread for 'ops' milliseconds.
*/
int WaitTest(LNG_TYPE ops, LNG_TYPE dummy, const void* buf)
{
	::Sleep(ops);
	return 0;
}


/**
	Performs integer operations.
	ops is the number of operations to perform
*/
int MeasureIOPS( LNG_TYPE ops, LNG_TYPE dummy, const void* buf )
{
	while( ops > 0 )
	{
		ops = ops - 1;
	}                  
	         
	return 0;
}

/**
	Performs floating point operations
	ops is the number of operations to perform
*/
int MeasureFLOPS( LNG_TYPE iops, LNG_TYPE dummy, const void* buf )
{
	FP_TYPE ops = iops;
	assert (ops == iops);		// If this fails, we have a too fast machine or too small FP_TYPE

	while( ops > 0 )
	{
		ops = ops - 1.00000000001;
	}
	
	return 0;
}

/**
	Performs bit shifting
	ops is the number of operations to perform
*/
int MeasureBitShifting( LNG_TYPE ops, LNG_TYPE dummy, const void* buf )
{
	LNG_TYPE foo = 0xFFFFFFFF;

	LNG_TYPE startFoo  = foo;

	ops /= 10;	// we do 10 operations at once
     
	while( ops --> 0 )
	{         
		foo = foo<<1;
		foo = foo>>1;
		foo = foo<<2;
		foo = foo>>2;
		foo = foo<<3;

		foo = foo>>3;
		foo = foo<<4;
		foo = foo>>4;
		foo = foo<<2;
		foo = foo>>2;
	}

	// Sanity check... :)
	if(startFoo == foo)
		return 0; 

	return -1;
}

/**
	Performs mathmatical floating point calculations.
	ops defines how many calculations shall be done.
*/
int MeasureFloatingMath( LNG_TYPE ops, LNG_TYPE v, const void* buf )
{
	FP_TYPE a, b, c, d, e, vv;
	vv = v;

	ops /= 5;	// we do 5 operations at once
	double i;

	while( ops --> 0 )
	{
		a = sqrt( vv );
		b = log( vv );
		c = pow( a, 2 );
		d = exp( b );
		e = modf( a , &i );
	}

	return 0;
}

/**
	Estimates the raw speed of memory ops
	ops = number of operations to perform, expected to be a multiple of 5
	b is the desired block size in KiB
*/
int ApproximateMemCpy(LNG_TYPE ops, LNG_TYPE b, const void* buf)
{
	size_t bufSize = b * 1024;		// KiB

	void* buf1 = new char[bufSize]; // malloc(bufSize);
	
	if( buf1 == NULL )
	{
		return -1;     
	}

	void* buf2 = new char[bufSize]; // malloc(bufSize);
	if (buf2 == NULL)
	{
		free( buf1 );
		return -1;
	}

	LNG_TYPE c = ops/5;
	while(c --> 0)
	{		
		MEMCPY(buf1, buf2, bufSize); 
		MEMCPY(buf2, buf1, bufSize);
		MEMCPY(buf1, buf2, bufSize);
		MEMCPY(buf2, buf1, bufSize);
		MEMCPY(buf1, buf2, bufSize);
	}		
	
	delete buf1;
	delete buf2;

	return 0;
}

/**
	Estimates the raw speed of memory ops
	ops = number of operations to perform, expected to be a multiple of 5
	b = bytes to set per operation
*/
int ApproximateMemSet(LNG_TYPE ops, LNG_TYPE b, const void* buf)
{
	size_t bufSize = b;		

	char* buf1 = new char[bufSize]; // malloc(bufSize);
	if (buf1 == NULL)
	{
		return -1;
	}

	LNG_TYPE c = ops/5;
	while (c --> 0)
	{
		MEMSET(buf1, 0xFE, b);
		MEMSET(buf1, 0xFE, b);
		MEMSET(buf1, 0xFE, b);
		MEMSET(buf1, 0xFE, b);
		MEMSET(buf1, 0xFE, b);
	}

	delete buf1;

	return 0;
}

/**
	Performs 'ops' MiB disk write operations.
	ops is the number of operations to perform
	b is the desired block size in KiB
	buf is the path used for testing.
*/
int ApproximateDiskWriteSpeed( LNG_TYPE ops, LNG_TYPE b, const void* buf )
{
	char* szPath = (char*)buf;
	if( szPath == NULL )
		return -1;

	FILE *fp = fopen( szPath, "wbc" );		// write binary, write through
	if( fp == NULL )
	{
	 	return -1;
	}	

	LNG_TYPE c = ops;

	/// *******

#ifdef _WIN32

	LNG_TYPE blockSize = b * 1024;	// KiB
	void* buf1 = new char[blockSize]; // malloc(blockSize);
	if( buf1 == NULL )
		return -1;     
	
	while(c --> 0)
	{		
		fwrite( buf1, 1, blockSize, fp );
		fseek( fp, 0, SEEK_SET );				// Rewind ptr to keep it small 
		fflush( fp );	// Flush to disk
	}
	
	delete buf1;
		
#else // _DOS _WIN16

	// This is a more portable version (no 16 bit limit)
	// but it is slower.
	LNG_TYPE blockSize = 1024;	// Byte
	void* buf1 = malloc(blockSize);
	if( buf1 == NULL )
		return -1;     

	LNG_TYPE bc = 0;
	while(c --> 0)
	{		
		bc = b;
		while( bc-->0 )
		{
			fwrite( buf1, 1, blockSize, fp );
		}
		fflush( fp );							// Flush to disk
		fseek( fp, 0, SEEK_SET );				// Rewind ptr to keep it small 
	}

	free( buf1 );

#endif 

	/// *******

	fclose( fp );
	remove( szPath );  

	return 0; 
}
